#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 00:13:48 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: allesfitter modules
from .basement import Basement




def init(datadir):
    global BASEMENT
    BASEMENT = Basement(datadir)


"""
Pipeline to analyze allesfitter output for planet transit timings

argument 1: allesfitter path
argument 2: p-value threshold
argument 3: Boolean to select to plot wout/with TESS or wout/with/only TESS

Tansu Daylan
MIT Kavli Institute, Cambridge, MA, 02109, US
tansu.daylan@gmail.com
www.tansudaylan.com
"""

import numpy as np
import scipy
import os, datetime, sys

import matplotlib.pyplot as plt

import allesfitter
import allesfitter.postprocessing.plot_viol
from allesfitter import config

import astropy

def plot_viol(pathdataoutp, pvalthrs=1e-3, boolonlytess=False):

    strgtimestmp = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    print('allesfitter postprocessing violin plot started at %s...' % strgtimestmp)
    
    liststrgruns = ['woutTESS', 'alldata']
    if boolonlytess:
        liststrgruns.append(['TESS'])
    
    numbruns = len(liststrgruns)
    indxruns = np.arange(numbruns)
    
    liststrgpara = [[] for i in indxruns]
    listlablpara = [[] for i in indxruns]
    listobjtalle = [[] for i in indxruns]
    for i, strgruns in enumerate(liststrgruns):
        pathdata = pathdataoutp + 'allesfits/allesfit_%s' % strgruns
        print('Reading from %s...' % pathdata)
        config.init(pathdata)
        liststrgpara[i] = np.array(config.BASEMENT.fitkeys)
        listlablpara[i] = np.array(config.BASEMENT.fitlabels)
        # read the chain
        listobjtalle[i] = allesfitter.allesclass(pathdata)
        
    liststrgparaconc = np.concatenate(liststrgpara)
    liststrgparaconc = np.unique(liststrgparaconc)
    listlablparaconc = np.copy(liststrgparaconc)
    for k, strgparaconc in enumerate(liststrgparaconc):
        for i, strgruns in enumerate(liststrgruns):
            if strgparaconc in liststrgpara[i]:
                listlablparaconc[k] = listlablpara[i][np.where(liststrgpara[i] == strgparaconc)[0][0]]
    
    ticklabl = ['w/o TESS', 'w/ TESS']
    if boolonlytess:
        ticklabl.append(['only TESS'])

    for k, strgpara in enumerate(liststrgparaconc):
        booltemp = True
        for i in indxruns:
            if not strgpara in liststrgpara[i]:
                booltemp = False
        if not booltemp:
            continue
        
        figr, axis = plt.subplots()
        chanlist = []
        for i in indxruns:
            chanlist.append((listobjtalle[i].posterior_params[strgpara] - np.mean(listobjtalle[i].posterior_params[strgpara])) * 24. * 60.)
        axis.violinplot(chanlist, showmedians=True, showextrema=False)
        axis.set_xticks(np.arange(numbruns) + 1)
        axis.set_xticklabels(ticklabl)
        if strgpara == 'b_period':
            axis.set_ylabel('P [min]')
        else:
            axis.set_ylabel(listlablparaconc[k])
        plt.tight_layout()
        
        path = pathdataoutp + 'viol_%04d.pdf' % (k)
        print('Writing to %s...' % path)
        figr.savefig(path)
        plt.close()
    
    listyear = [2021, 2023, 2025]
    numbyear = len(listyear)
    indxyear = np.arange(numbyear)
    timejwst = [[[] for i in indxruns] for k in indxyear]
    for k, year in enumerate(listyear):
        epocjwst = astropy.time.Time('%d-01-01 00:00:00' % year, format='iso').jd
        for i in indxruns:
            epoc = listobjtalle[i].posterior_params['b_epoch']
            peri = listobjtalle[i].posterior_params['b_period']
            indxtran = (epocjwst - epoc) / peri
            indxtran = np.mean(np.rint(indxtran))
            if indxtran.size != np.unique(indxtran).size:
                raise Exception('')

            timejwst[k][i] = epoc + peri * indxtran

            timejwst[k][i] -= np.mean(timejwst[k][i])
            timejwst[k][i] *= 24. * 60.
    
    listfigr = []
    listaxis = []

    ## temporal evolution
    figr, axis = plt.subplots(figsize=(12, 6))
    listfigr.append(figr)
    listaxis.append(axis)
    axis.violinplot([timejwst[k][1] for k in indxyear], listyear)
    axis.set_xlabel('Year')
    axis.set_ylabel('Transit time residual [min]')
    plt.tight_layout()
    path = pathdataoutp + 'jwsttime.pdf'
    print('Writing to %s...' % path)
    plt.savefig(path)
    plt.close()
    
    ## without/with/only TESS prediction comparison
    figr, axis = plt.subplots(figsize=(12, 6))
    listfigr.append(figr)
    listaxis.append(axis)
    axis.violinplot(timejwst[1], range(len(liststrgruns)), points=2000)
    axis.set_xticks(range(len(liststrgruns)))
    axis.set_xticklabels(ticklabl)
    axis.set_ylabel('Transit time residual in 2023 [min]')
    #axis.set_ylim([-300, 300])
    plt.tight_layout()
    path = pathdataoutp + 'jwstcomp.pdf'
    print('Writing to %s...' % path)
    plt.savefig(path)
    plt.close()
    
    return listfigr, listaxis

    # all parameter summary
    figr, axis = plt.subplots()
    chanlist = []
    axis.violinplot(chanlist, showmedians=True, showextrema=False)
    axis.set_xticks(valutick)
    axis.set_xticklabels(labltick)
    axis.set_ylabel(lablparatemp)
    plt.tight_layout()
    path = pathdataoutp + 'para_%s.pdf'
    print('Writing to %s...' % path)
    figr.savefig(path)
    plt.close()

    # plot p values
    ## threshold p value to conclude significant difference between posteriors with and without TESS
    if pvalthrs is None:
        pvalthrs = 1e-6
    
    lablparacomp = [[] for u in indxruns]
    for u in indxruns:
    
        lablparacomp[u] = list(set(lablpara[indxrunsfrst[u]]).intersection(lablpara[indxrunsseco[u]]))
    
        # post-processing
        ## calculate the KS test statistic between the posteriors
        numbparacomp = len(lablparacomp[u])
        pval = np.empty(numbparacomp)
        for j in range(numbparacomp):
            kosm, pval[j] = scipy.stats.ks_2samp([indxrunsfrst[u]][:, j], chan[indxrunsseco[u]][:, j])
            kosm, pval[j] = scipy.stats.ks_2samp(chan[indxrunsfrst[u]][:, j], chan[indxrunsseco[u]][:, j])
    
        ## find the list of parameters whose posterior with and without TESS are unlikely to be drawn from the same distribution
        indxparagood = np.where(pval < pvalthrs)[0]
        if indxparagood.size > 0:
    
            figr, axis = plt.subplots(figsize=(12, 5))
            indxparacomp = np.arange(numbparacomp)
            axis.plot(indxparacomp, pval, ls='', marker='o')
            axis.plot(indxparacomp[indxparagood], pval[indxparagood], ls='', marker='o', color='r')
            axis.set_yscale('log')
            axis.set_xticks(indxparacomp)
            axis.set_xticklabels(lablparacomp[u])
            if u == 0:
                axis.set_title('Posteriors with TESS vs. without TESS')
            if u == 1:
                axis.set_title('Posteriors without TESS vs. only TESS')
            if u == 2:
                axis.set_title('Posteriors with TESS vs. only TESS')
    
            axis.axhline(pvalthrs, ls='--', color='black', alpha=0.3)
            plt.tight_layout()
            path = pathdataoutp + 'kosm_com%d.pdf' % u
            print('Writing to %s...' % path)
            figr.savefig(path)
            plt.close()
    

    





# -*- coding: utf-8 -*-
"""
Created on Wed Apr 27 12:55:39 2016

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research,
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109,
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np

#::: lightcurves modules
from .utils import mask_ranges




def get_first_epoch(time, epoch, period, width=0):    
    '''
    width : float
        set >0 to include transit egress to mark the first transit
    place the first_epoch at the start of the data to avoid luser mistakes
    '''
    time = np.sort(time)
    start = np.nanmin( time )
    first_epoch = 1.*epoch + width/2. #add width/2 to catch egress
    if start<=first_epoch: first_epoch -= int(np.round((first_epoch-start)/period)) * period
    else: first_epoch += int(np.round((start-first_epoch)/period)) * period
    return first_epoch - width/2.  #subtract width/2 to get midpoint again
    
    

def index_transits(time, epoch, period, width):
    """
    Returns:
    --------
    ind_tr : array
        indices of points in transit
    ind_out : array
        indices of points out of transit
    """
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=width)
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
    
    tmid = np.array( [ epoch + i * period for i in range(N) ] )
    
    _, ind_tr, mask_tr = mask_ranges( time, tmid - width/2., tmid + width/2. )
    ind_out = np.arange( len(time) )[ ~mask_tr ]

    return ind_tr, ind_out 
    
    

#::: for binaries, mark the primary and secondary eclipse
#TODO: implement non-circular orbits
def index_eclipses(time, epoch, period, width_1, width_2):
    """
    Returns:
    --------
    ind_ecl1 : array
        indices of points in primary eclipse
    ind_ecl2 : array
        indices of points in secondary eclipse
    ind_out : array
        outside of any eclipse
    
    ! this assumes circular orbits !
    """
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=width_1)
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
        
    tmid_ecl1 = np.array( [ epoch +             i * period  for i in range(N) ] )
    tmid_ecl2 = np.array( [ epoch - period/2. + i * period  for i in range(N+1) ] )
    
    _, ind_ecl1, mask_ecl1 = mask_ranges( time, tmid_ecl1 - width_1/2., tmid_ecl1 + width_1/2. )           
    _, ind_ecl2, mask_ecl2 = mask_ranges( time, tmid_ecl2 - width_2/2., tmid_ecl2 + width_2/2. )
        
    ind_out = np.arange( len(time) )[ ~(mask_ecl1 | mask_ecl2) ]

    return ind_ecl1, ind_ecl2, ind_out
    


def get_tmid_transits(time, epoch, period, width):
    '''
    get a list of only the transit midpoints that are actually covered by the data
    '''
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=width)
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
    tmid = np.array( [ epoch + i * period for i in range(N) ] )
    
    return tmid


    
def get_tmid_observed_transits(time, epoch, period, width):
    '''
    get a list of only the transit midpoints that are actually covered by the data
    '''
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=width)
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
    tmid = np.array( [ epoch + i * period for i in range(N) ] )
    
    tmid_observed_transits = []
    
    for i,t in enumerate(tmid): 
        mask = ((time >= (t - width/2.)) & (time <= (t + width/2.)))
        if any(mask): 
            tmid_observed_transits.append( tmid[i] )
    
    return tmid_observed_transits




#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 01:05:28 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import os
import dynesty
from scipy.special import ndtri
from scipy.stats import truncnorm
from multiprocessing import Pool
from contextlib import closing
#import bzip2
import gzip
try:
   import cPickle as pickle
except:
   import pickle
from time import time as timer

#::: warnings
import warnings
warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning) 
warnings.filterwarnings('ignore', category=np.RankWarning) 

#::: allesfitter modules
from . import config
from .computer import update_params, calculate_lnlike_total
from .general_output import logprint




###############################################################################
#::: Nested Sampling log likelihood
###############################################################################
def ns_lnlike(theta):
    
    params = update_params(theta)
    lnlike = calculate_lnlike_total(params)
    
#    lnlike = 0
#    
#    for inst in config.BASEMENT.settings['inst_phot']:
#        lnlike += calculate_lnlike(params, inst, 'flux')
#    
#    for inst in config.BASEMENT.settings['inst_rv']:
#        lnlike += calculate_lnlike(params, inst, 'rv')
#        
#    if np.isnan(lnlike) or np.isinf(lnlike):
#        lnlike = -np.inf
        
    return lnlike



###############################################################################
#::: Nested Sampling prior transform
###############################################################################
def ns_prior_transform(utheta):
#    global config.BASEMENT
    theta = np.zeros_like(utheta)*np.nan
    for i in range(len(theta)):
        if config.BASEMENT.bounds[i][0]=='uniform':
            theta[i] = utheta[i]*(config.BASEMENT.bounds[i][2]-config.BASEMENT.bounds[i][1]) + config.BASEMENT.bounds[i][1]
        elif config.BASEMENT.bounds[i][0]=='normal':
            theta[i] = config.BASEMENT.bounds[i][1] + config.BASEMENT.bounds[i][2]*ndtri(utheta[i])
        elif config.BASEMENT.bounds[i][0]=='trunc_normal':
            theta[i] = my_truncnorm_isf(utheta[i],config.BASEMENT.bounds[i][1],config.BASEMENT.bounds[i][2],config.BASEMENT.bounds[i][3],config.BASEMENT.bounds[i][4]) 
        else:
            raise ValueError('Bounds have to be "uniform", "normal" and "trunc_normal". Input from "params.csv" was "'+config.BASEMENT.bounds[i][0]+'".')
    return theta
    

def my_truncnorm_isf(q,a,b,mean,std):
    a_scipy = 1.*(a - mean) / std
    b_scipy = 1.*(b - mean) / std
    return truncnorm.isf(q,a_scipy,b_scipy,loc=mean,scale=std)



###############################################################################
#::: Nested Sampling fitter class
###############################################################################
def ns_fit(datadir):
    
    #::: init
    config.init(datadir)
        
        
    #::: settings
    nlive  = config.BASEMENT.settings['ns_nlive']    # (default 500) number of live points
    bound  = config.BASEMENT.settings['ns_bound']    # (default 'single') use MutliNest algorithm for bounds
    ndim   = config.BASEMENT.ndim                    # number of parameters
    sample = config.BASEMENT.settings['ns_sample']   # (default 'auto') random walk sampling
    tol    = config.BASEMENT.settings['ns_tol']      # (defualt 0.01) the stopping criterion
        
     
    #::: run
    if config.BASEMENT.settings['ns_modus']=='static':
        logprint('\nRunning Static Nested Sampler...')
        logprint('--------------------------')
        t0 = timer()
        
        if config.BASEMENT.settings['multiprocess']:
             with closing(Pool(processes=(config.BASEMENT.settings['multiprocess_cores']))) as pool:
                logprint('\nRunning on', config.BASEMENT.settings['multiprocess_cores'], 'CPUs.')
                sampler = dynesty.NestedSampler(ns_lnlike, ns_prior_transform, ndim, 
                                                pool=pool, queue_size=config.BASEMENT.settings['multiprocess_cores'], 
                                                bound=bound, sample=sample, nlive=nlive)
                sampler.run_nested(dlogz=tol, print_progress=config.BASEMENT.settings['print_progress'])
            
        else:
            sampler = dynesty.NestedSampler(ns_lnlike, ns_prior_transform, ndim,
                                            bound=bound, sample=sample, nlive=nlive)
            sampler.run_nested(dlogz=tol, print_progress=config.BASEMENT.settings['print_progress'])
            
        t1 = timer()
        timedynesty = (t1-t0)
        logprint("\nTime taken to run 'dynesty' (in static mode) is {} hours".format(int(timedynesty/60./60.)))


    elif config.BASEMENT.settings['ns_modus']=='dynamic':
        logprint('\nRunning Dynamic Nested Sampler...')
        logprint('--------------------------')
        t0 = timer()
        
        if config.BASEMENT.settings['multiprocess']:
             with closing(Pool(processes=config.BASEMENT.settings['multiprocess_cores'])) as pool:
                logprint('\nRunning on', config.BASEMENT.settings['multiprocess_cores'], 'CPUs.')
                sampler = dynesty.DynamicNestedSampler(ns_lnlike, ns_prior_transform, ndim, 
                                                       pool=pool, queue_size=config.BASEMENT.settings['multiprocess_cores'], 
                                                       bound=bound, sample=sample)
                sampler.run_nested(nlive_init=nlive, dlogz_init=tol, print_progress=config.BASEMENT.settings['print_progress'])
            
        else:
            sampler = dynesty.DynamicNestedSampler(ns_lnlike, ns_prior_transform, ndim,
                                                   bound=bound, sample=sample)
            sampler.run_nested(nlive_init=nlive, print_progress=config.BASEMENT.settings['print_progress'])
            
        t1 = timer()
        timedynestydynamic = (t1-t0)
        logprint("\nTime taken to run 'dynesty' (in dynamic mode) is {:.2f} hours".format(timedynestydynamic/60./60.))


    #::: pickle-save the 'results' class
    results = sampler.results
    f = gzip.GzipFile(os.path.join(config.BASEMENT.outdir,'save_ns.pickle.gz'), 'wb')
    pickle.dump(results, f)
    f.close()


    


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Credit & source: https://bitbucket.org/william_rusnack/to-precision/src/master/
"""


from __future__ import print_function, division, absolute_import


__author__ = '''William Rusnack github.com/BebeSparkelSparkel linkedin.com/in/williamrusnack williamrusnack@gmail.com
Eric Moyer github.com/epmoyer eric@lemoncrab.com'''



from math import floor, log10


def std_notation(value, precision):
  '''
  standard notation (US version)
  ref: http://www.mathsisfun.com/definitions/standard-notation.html

  returns a string of value with the proper precision

  ex:
    std_notation(5, 2) => '5.0'
    std_notation(5.36, 2) => '5.4'
    std_notation(5360, 2) => '5400'
    std_notation(0.05363, 3) => '0.0536'

    created by William Rusnack
      github.com/BebeSparkelSparkel
      linkedin.com/in/williamrusnack/
      williamrusnack@gmail.com
  '''
  return to_precision(value, precision, notation='std')


def sci_notation(value, precision, delimiter='e'):
  '''
  scientific notation
  ref: https://www.mathsisfun.com/numbers/scientific-notation.html

  returns a string of value with the proper precision and 10s exponent
  delimiter is placed between the decimal value and 10s exponent

  ex:
    sci_notation(123, 1, 'E') => '1E2'
    sci_notation(123, 3, 'E') => '1.23E2'
    sci_notation(.126, 2, 'E') => '1.3E-1'

    created by William Rusnack
      github.com/BebeSparkelSparkel
      linkedin.com/in/williamrusnack/
      williamrusnack@gmail.com
  '''
  return to_precision(value, precision, notation='sci', delimiter=delimiter)


def eng_notation(value, precision, delimiter='e'):
  '''
  engineering notation
  ref: http://www.mathsisfun.com/definitions/engineering-notation.html

  returns a string of value with the proper precision and 10s exponent that is divisible by 3
  delimiter is placed between the decimal value and 10s exponent

  ex:
    sci_notation(123, 1, 'E') => '100E0'
    sci_notation(1230, 3, 'E') => '1.23E3'
    sci_notation(.126, 2, 'E') => '120E-3'

  created by William Rusnack
    github.com/BebeSparkelSparkel
    linkedin.com/in/williamrusnack/
    williamrusnack@gmail.com
  '''
  return to_precision(value, precision, notation='eng', delimiter=delimiter)


def auto_notation(value, precision, delimiter='e'):
  '''
  Automatically selects between standard notation (US version) and scientific notation.
  Values in the range 0.001 < abs(value) < 1000 return standard notation.

  http://www.mathsisfun.com/definitions/standard-notation.html
  https://www.mathsisfun.com/numbers/scientific-notation.html

  returns a string of value with the proper precision

  ex:
    auto_notation(123, 4) => '123.4'
    std_notation(1234, 4) => '1.234e3'
  '''
  return to_precision(value, precision, notation='auto', delimiter=delimiter)


def to_precision(
  value,
  precision,
  notation='auto',
  delimiter='e',
  auto_limit=3,
  strip_zeros=False,
  preserve_integer=False):
  '''
  converts a value to the specified notation and precision
  value - any type that can be converted to a float
  precision - integer that is greater than zero
  notation - string
    'auto' - selects standard notation when abs(power) < auto_limit else
      returns scientific notation.
    'sci' or 'scientific' - returns scientific notation
      ref: https://www.mathsisfun.com/numbers/scientific-notation.html
    'eng' or 'engineering' - returns engineering notation
      ref: http://www.mathsisfun.com/definitions/engineering-notation.html
    'std' or 'standard' - returns standard notation
      ref: http://www.mathsisfun.com/definitions/standard-notation.html
  delimiter - is placed between the decimal value and 10s exponent
  auto_limit - integer. When abs(power) exceeds this limit, 'auto'
    mode will return scientific notation.
  strip_zeros - if true, trailing decimal zeros will be removed.
  preserve_integer - if true, 'std' will preserve all digits when returning
    values that have no decimal component.
  '''
  is_neg, sig_digits, dot_power, ten_power = _sci_decompose(value, precision)

  if notation == 'auto':
    if abs(ten_power) < auto_limit:
      converter = _std_notation
    else:
      converter = _sci_notation

  elif notation in ('sci', 'scientific'):
    converter = _sci_notation

  elif notation in ('eng', 'engineering'):
    converter = _eng_notation

  elif notation in ('std', 'standard'):
    converter = _std_notation

  else:
    raise ValueError('Unknown notation: ' + str(notation))

  return converter(value, precision, delimiter, strip_zeros, preserve_integer)


def _std_notation(value, precision, _, strip_zeros, preserve_integer):
  '''
  standard notation (US version)
  ref: http://www.mathsisfun.com/definitions/standard-notation.html

  returns a string of value with the proper precision

  strip_zeros - if true, trailing decimal zeros will be removed.
  preserve_integer - if true, 'std' will preserve all digits when returning
    values that have no decimal component.
  '''
  sig_digits, power, is_neg = _number_profile(value, precision)
  result = ('-' if is_neg else '') + _place_dot(sig_digits, power, strip_zeros)

  if preserve_integer and not '.' in result:
    # Result was an integer, preserve all digits
    result = '{:0.0f}'.format(value)

  return result


def _sci_notation(value, precision, delimiter, strip_zeros, _):
  '''
  scientific notation
  ref: https://www.mathsisfun.com/numbers/scientific-notation.html

  returns a string of value with the proper precision and 10s exponent
  delimiter is placed between the decimal value and 10s exponent

  strip_zeros - if true, trailing decimal zeros will be removed.
  '''
  is_neg, sig_digits, dot_power, ten_power = _sci_decompose(value, precision)

  return ('-' if is_neg else '') + _place_dot(sig_digits, dot_power, strip_zeros) + delimiter + str(ten_power)


def _eng_notation(value, precision, delimiter, strip_zeros, _):
  '''
  engineering notation
  ref: http://www.mathsisfun.com/definitions/engineering-notation.html

  returns a string of value with the proper precision and 10s exponent that is divisible by 3
  delimiter is placed between the decimal value and 10s exponent

  strip_zeros - if true, trailing decimal zeros will be removed.
  '''
  is_neg, sig_digits, dot_power, ten_power = _sci_decompose(value, precision)

  eng_power = int(3 * floor(ten_power / 3))
  eng_dot = dot_power + ten_power - eng_power

  return ('-' if is_neg else '') + _place_dot(sig_digits, eng_dot, strip_zeros) + delimiter + str(eng_power)


def _sci_decompose(value, precision):
  '''
  returns the properties for to construct a scientific notation number
  used in sci_notation and eng_notation

  created by William Rusnack
    github.com/BebeSparkelSparkel
    linkedin.com/in/williamrusnack/
    williamrusnack@gmail.com
  '''
  value = float(value)
  sig_digits, power, is_neg = _number_profile(value, precision)

  dot_power = -(precision - 1)
  ten_power = power + precision - 1

  return is_neg, sig_digits, dot_power, ten_power


def _place_dot(digits, power, strip_zeros=False):
  '''
  places the dot in the correct spot in the digits
  if the dot is outside the range of the digits zeros will be added
  if strip_zeros is set, trailing decimal zeros will be removed

  ex:
    _place_dot('123',   2, False) => '12300'
    _place_dot('123',  -2, False) => '1.23'
    _place_dot('123',   3, False) => '0.123'
    _place_dot('123',   5, False) => '0.00123'
    _place_dot('120',   0, False) => '120.'
    _place_dot('1200', -2, False) => '12.00'
    _place_dot('1200', -2, True ) => '12'
    _place_dot('1200', -1, False) => '120.0'
    _place_dot('1200', -1, True ) => '120'

    created by William Rusnack
      github.com/BebeSparkelSparkel
      linkedin.com/in/williamrusnack/
      williamrusnack@gmail.com
  '''
  if power > 0:
    out = digits + '0' * power

  elif power < 0:
    power = abs(power)
    precision = len(digits)

    if power < precision:
      out = digits[:-power] + '.' + digits[-power:]

    else:
      out = '0.' + '0' * (power - precision) + digits

  else:
    out = digits + ('.' if digits[-1] == '0' and len(digits) > 1 else '')

  if strip_zeros and '.' in out:
    out = out.rstrip('0').rstrip('.')

  return out


def _number_profile(value, precision):
  '''
  returns:
    string of significant digits
    10s exponent to get the dot to the proper location in the significant digits
    bool that's true if value is less than zero else false

    created by William Rusnack
      github.com/BebeSparkelSparkel
      linkedin.com/in/williamrusnack/
      williamrusnack@gmail.com
  '''
  value = float(value)
  if value == 0:
    sig_digits = '0' * precision
    power = -(1 - precision)
    is_neg = False

  else:
    if value < 0:
      value = abs(value)
      is_neg = True
    else:
      is_neg = False

    power = -1 * floor(log10(value)) + precision - 1
    sig_digits = str(int(round(abs(value) * 10.0**power)))

  return sig_digits, int(-power), is_neg

